/* MoreBalance includes */
#include "conf.h"
#include "server.h"
#include "spec.h"

/* System includes */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/wait.h>
#include <getopt.h>

/* Definitions */
#define VERSION "0.2"

/* Global configuration */
static char *configfile = NULL;
static char *configstring = NULL;
static int verbosity = 1;

/* Child function: startup MoreBalance */
static void launch()
{
	conf c;

	if(configstring)
		c = conf_readstring(configstring);
	else
		c = conf_read(configfile);

	if(c.valid)
	{
		if(verbosity)
		{
			printf("Configuration loaded.\n");
			fflush(NULL);
		}

		conf_optimize(&c);
		if(verbosity) conf_dump(c);
		if(verbosity) conf_check(c);

		if(verbosity)
		{
			printf("Start the server.\n");
			fflush(NULL);
		}

		server_start(c);
	}
	else
	{
		if(configstring)
			fprintf(stderr, "Error reading configuration commands.\n");
		else
			fprintf(stderr, "Error reading configuration file.\n");
		exit(-1);
	}
}

/* Available options */
static struct option const options[] =
{
	{"configuration", required_argument, NULL, 'c'},
	{"daemon", no_argument, NULL, 'd'},
	{"help", no_argument, NULL, 'h'},
	{"immediate", required_argument, NULL, 'i'},
	{"spec", no_argument, NULL, 's'},
	{"version", no_argument, NULL, 'v'},
	{NULL, 0, NULL, 0}
};

/* Main function: main safety loop */
int main(int argc, char *argv[])
{
	pid_t pid;
	int status;
	int daemonmode;
	int c, i;

	daemonmode = 0;

	while((c = getopt_long(argc, argv, "c:dhi:sv", options, NULL)) != -1)
	{
		switch(c)
		{
			case 'c':
				configfile = optarg;
				break;
			case 'd':
				daemonmode = 1;
				verbosity = 0;
				break;
			case 'h':
				printf("MoreBalance - a sophisticated load balancing system\n");
				printf("See http://mindx.dyndns.org/morebalance/ for information\n");
				printf("\n");
				printf("Options:\n");
				printf("[-c | --configuration <file>] Use configuration from <file>\n");
				printf("[-i | --immediate <string>  ] Use command line configuration\n");
				printf("[-d | --daemon              ] Run as daemon in the background\n");
				printf("\n");
				printf("[-h | --help                ] Show this help\n");
				printf("[-v | --version             ] Display version information\n");
				printf("[-s | --spec                ] Display build specification\n");
				exit(0);
				break;
			case 'i':
				configstring = optarg;
				break;
			case 's':
				printf("Build specifications\n");
				printf("TLS support:  ");
#ifdef USE_TLS
				printf("yes\n");
#else
				printf("no\n");
#endif
				printf("PCRE support: ");
#ifdef USE_PCRE
				printf("yes\n");
#else
				printf("no\n");
#endif
				exit(0);
				break;
			case 'v':
				printf("MoreBalance %s\n", VERSION);
				exit(0);
				break;
			default:
				fprintf(stderr, "Unknown option, try --help.\n");
				exit(-1);
		}
	}

	for(i = optind; i < argc; i++)
	{
		configstring = argv[i];
	}

	if((!configfile) && (!configstring)) configfile = "morebalance.conf";

	if(verbosity)
	{
		printf(">> Startup.\n");
		fflush(NULL);
	}

	if(daemonmode)
	{
		pid = fork();
		if(pid > 0) exit(0);
		else if(pid < 0)
		{
			fprintf(stderr, "Couldn't daemonize,\n");
			exit(-1);
		}
		if(verbosity)
		{
			printf(">> Running as daemon\n");
			fflush(NULL);
		}
	}

	while(1)
	{
		pid = fork();
		switch(pid)
		{
			case -1:
				fprintf(stderr, "Couldn't fork.\n");
				exit(-1);
				break;
			case 0:
				launch();
				exit(-1);
				break;
			default:
				break;
		}

		if(verbosity)
		{
			printf(">> Loop.\n");
			fflush(NULL);
		}

		wait(&status);
		if(!WIFSIGNALED(status)) break;

		fprintf(stderr, ">> Child crashed. Restarting.\n");
	}

	return 0;
}

