/* MoreBalance includes */
#include "server.h"
#include "magic.h"
#include "ipban.h"
#include "sys.h"
#include "filter.h"
#include "spec.h"

/* System includes */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>
#include <netinet/in.h>
#include <pthread.h>
#include <netdb.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#ifdef USE_TLS
#include <ggz.h>
#endif

/* Definitions */
#define FILTER_MAX 512
#define BUFFER_MAX 16384

/* Globals */
static conf global_c;

/* Connection structure */
struct conn_t
{
	int fd;
	char *hostname;
	char *prelaunch;
	char *pattern;
	int port;
	int portopts;
	char *socket;
	int destport;
	int destportopts;
	char *destsocket;
};

/* Sub-subserver for each connection */
static void *server_handle(void *arg)
{
	struct sockaddr_in sa;
	struct sockaddr_un su;
	struct hostent *h = NULL;
	int fd, fd2;
	int ret, ret2;
	ssize_t bytes, tmpbytes;
	char buf[BUFFER_MAX];
	char filter[FILTER_MAX];
	long long rbytes, sbytes, delaybytes;
	struct conn_t *carg;
	pid_t pid;

	ret = 1;

	carg = (struct conn_t*)arg;
	fd = carg->fd;

	if((carg->hostname[0] != '/') && (carg->destport != 0))
	{
		h = gethostbyname(carg->hostname);
		if(!h)
		{
			fprintf(stderr, "Couldn't resolve %s.\n", carg->hostname);

			magic_disable(carg->hostname, carg->destport);

			close(fd);
			free(carg);
			pthread_exit(&ret);
		}

		fd2 = socket(AF_INET, SOCK_STREAM, 0);
	}
	else
	{
		fd2 = socket(AF_LOCAL, SOCK_STREAM, 0);
	}

	if(fd2 < 0)
	{
		fprintf(stderr, "Couldn't create socket.\n");
		free(carg);
		pthread_exit(&ret);
	}

	if((carg->hostname[0] != '/') && (carg->destport != 0))
	{
		sa.sin_family = AF_INET;
		sa.sin_port = htons(carg->destport);
		memcpy(&sa.sin_addr, h->h_addr, h->h_length);
	}
	else
	{
		su.sun_family = AF_LOCAL;
		if(carg->hostname[0] == '/')
		{
			strncpy(su.sun_path, carg->hostname, (sizeof(su.sun_path) - 1));
			/*unlink(carg->hostname);*/
		}
		else
		{
			strncpy(su.sun_path, carg->destsocket, (sizeof(su.sun_path) - 1));
			/*unlink(carg->destsocket);*/
		}
	}

	if((carg->hostname[0] != '/') && (carg->destport != 0))
		ret2 = connect(fd2, (const struct sockaddr*)&sa, sizeof(sa));
	else
		ret2 = connect(fd2, (const struct sockaddr*)&su, sizeof(su));

	if(ret2 < 0)
	{
		if(carg->destport)
			fprintf(stderr, "Couldn't connect to %s:%i.\n", carg->hostname, carg->destport);
		else
			fprintf(stderr, "Couldn't connect to %s:%s.\n", carg->hostname, carg->destsocket);
		if(carg->prelaunch)
		{
			fprintf(stderr, "Launch program: %s.\n", carg->prelaunch);
			pid = fork();
			switch(pid)
			{
				case -1:
					/* bail out anyway */
					fprintf(stderr, "Launch error.\n");
					break;
				case 0:
					execl(carg->prelaunch, carg->prelaunch, NULL);
					_exit(0);
					break;
				default:
					sleep(7);

					if((carg->hostname[0] != '/') && (carg->destport != 0))
						ret2 = connect(fd2, (const struct sockaddr*)&sa, sizeof(sa));
					else
						ret2 = connect(fd2, (const struct sockaddr*)&su, sizeof(su));

					if(ret2 < 0)
					{
						fprintf(stderr, "Port or socket still not available.\n");
					}
			}
		}
	}

	if(ret2 < 0)
	{
		if((carg->hostname[0] != '/') && (carg->destport != 0))
			magic_disable(carg->hostname, carg->destport);

		close(fd);
		free(carg);
		pthread_exit(&ret);
	}

#ifdef USE_TLS
	if(carg->destportopts & opt_tls)
	{
		ret2 = ggz_tls_enable_fd(fd2, GGZ_TLS_CLIENT, GGZ_TLS_VERIFY_NONE);
		if(!ret2) fprintf(stderr, "TLS handshake failed, continuing without encryption!\n");
	}
#endif

	fcntl(fd, F_SETFL, O_NONBLOCK);
	fcntl(fd2, F_SETFL, O_NONBLOCK);

	rbytes = 0;
	sbytes = 0;
	delaybytes = 0;

	if(global_c.verbose) printf("--> Transmission starting: fd=%i, fd2=%i!\n", fd, fd2);
	fflush(NULL);
	while(1)
	{
		bytes = read(fd, buf, sizeof(buf));
		if(bytes == -1)
		{
			if(errno != EAGAIN)
			{
				fprintf(stderr, "Read error on fd (client)!\n");
				break;
			}
		}
		else if(bytes > 0)
		{
			buf[bytes] = 0;
			if(global_c.verbose > 1) printf(":: -> %s", buf);
			fflush(NULL);
			if(carg->pattern)
			{
				tmpbytes = FILTER_MAX - delaybytes;
				if(tmpbytes < 0) tmpbytes = 0;
				else if(tmpbytes > bytes) tmpbytes = bytes;
				memcpy(filter + delaybytes, buf, tmpbytes);
				ret2 = filter_tellbinary(filter, delaybytes + tmpbytes);
				if(!ret2)
				{
					/*fprintf(stderr, "Doesn't match!\n");*/
					delaybytes += bytes;
					if(delaybytes + tmpbytes >= FILTER_MAX)
					{
						fprintf(stderr, "Match required but not detected!\n");
						break;
					}
					else continue;
				}
				else
				{
					/*printf("matched!\n");*/
					free(carg->pattern);
					carg->pattern = NULL;
					while(delaybytes)
					{
						tmpbytes = delaybytes - BUFFER_MAX;
						if(tmpbytes <= 0) tmpbytes = delaybytes;

						fprintf(stderr, "Delayed %qi bytes, resending %i.\n", delaybytes, tmpbytes);
						
						tmpbytes = write(fd2, filter + rbytes, tmpbytes);
						if(tmpbytes == -1)
						{
							if(errno != EAGAIN)
							{
								fprintf(stderr, "Write error on fd2 (node)!\n");
								break;
							}
						}
						rbytes += tmpbytes;
						delaybytes -= tmpbytes;
					}
					/*continue;*/
				}
			}

			bytes = write(fd2, buf, bytes);
			if(bytes == -1)
			{
				if(errno != EAGAIN)
				{
					fprintf(stderr, "Write error on fd2 (node)!\n");
					break;
				}
			}
			rbytes += bytes;
		}
		else if(bytes == 0)
		{
			if(global_c.verbose) fprintf(stderr, "broken pipe on fd (client)!\n");
			if(global_c.verbose > 1) printf("broken pipe on fd (client)!\n");
			fflush(NULL);
			break;
		}

		bytes = read(fd2, buf, sizeof(buf));
		if(bytes == -1)
		{
			if(errno != EAGAIN)
			{
				fprintf(stderr, "Read error on fd2 (node)!\n");
				break;
			}
		}
		else if(bytes > 0)
		{
			buf[bytes] = 0;
			if(global_c.verbose > 1) printf(":: <- %s", buf);
			fflush(NULL);
			bytes = write(fd, buf, bytes);
			if(bytes == -1)
			{
				if(errno != EAGAIN)
				{
					fprintf(stderr, "Write error on fd (client)!\n");
					break;
				}
			}
			sbytes += bytes;
		}
		else if(bytes == 0)
		{
			if(global_c.verbose) fprintf(stderr, "broken pipe on fd2 (cluster node)!\n");
			if(global_c.verbose > 1) printf("broken pipe on fd2 (cluster node)!\n");
			fflush(NULL);
			break;
		}
	}

	close(fd);
	close(fd2);

	if(global_c.verbose)
	{
		printf("Sent bytes [cluster node -> client]: %qi\n", sbytes);
		printf("Read bytes [client -> cluster node]: %qi\n", rbytes);
	}
	fflush(NULL);

	magic_traffic(&global_c, carg->hostname, sbytes, rbytes);

	if(carg->prelaunch) free(carg->prelaunch);
	if(carg->pattern) free(carg->pattern);
	if(carg->hostname) free(carg->hostname);
	free(carg);

	pthread_exit(&ret);
}

/* Threaded subserver */
static void *server_sub(void *arg)
{
	int ret, ret2;
	int fd;
	int port, destport, portopts, destportopts;
	struct sockaddr_in sa;
	struct sockaddr_un su;
	struct sockaddr addr;
	int addrsize;
	char *host;
	char *pattern, *prelaunch;
	char *portsocket, *destsocket;
	int x;
	struct conn_t *carg, *conn;
	pthread_t id;
	pthread_attr_t attr;
	char ip[15];
	unsigned int ipx[5];
	unsigned int local[5];

	ret = 1;

	conn = (struct conn_t*)arg;
	port = conn->port;
	portopts = conn->portopts;
	destportopts = conn->destportopts;
	pattern = conn->pattern;
	prelaunch = conn->prelaunch;
	portsocket = conn->socket;
	destsocket = conn->destsocket;
	free(conn);

	if(global_c.verbose) printf("=> Port: %i\n", port);
	fflush(NULL);

	if(portopts & opt_udp)
	{
		if(portsocket)
			fd = socket(AF_LOCAL, SOCK_DGRAM, 0);
		else
			fd = socket(AF_INET, SOCK_DGRAM, 0);
	}
	else
	{
		if(portsocket)
			fd = socket(AF_LOCAL, SOCK_STREAM, 0);
		else
			fd = socket(AF_INET, SOCK_STREAM, 0);
	}
	if(fd < 0)
	{
		fprintf(stderr, "Couldn't create file descriptor.\n");
		pthread_exit(&ret);
	}

	if(!(portopts & opt_udp))
	{
		setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &x, sizeof(x));
	}

	if(portsocket)
		su.sun_family = AF_LOCAL;
	else
	{
		sa.sin_family = AF_INET;
		if(portopts & opt_local)
		{
			local[1] = 127;
			local[2] = 0;
			local[3] = 0;
			local[4] = 1;
			local[0] = (local[1] << 24) + (local[2] << 16) + (local[3] << 8) + local[4];
			sa.sin_addr.s_addr = htonl(local[0]);
		}
		else
			sa.sin_addr.s_addr = htonl(INADDR_ANY);
	}
	if(portsocket)
		strncpy(su.sun_path, portsocket, (sizeof(su.sun_path) - 1));
	else
		sa.sin_port = htons(port);
	if(portsocket)
		ret = bind(fd, (const struct sockaddr*)&su, sizeof(su));
	else
		ret = bind(fd, (const struct sockaddr*)&sa, sizeof(sa));
	if(ret)
	{
		if(portsocket)
			fprintf(stderr, "Not all sockets are available.\n");
		else
			fprintf(stderr, "Not all ports are available.\n");
		sleep(1);
		pthread_exit(&ret);
	}

	if(!(portopts & opt_udp))
	{
		ret = listen(fd, 16);
		if(ret)
		{
			fprintf(stderr, "Couldn't listen to socket.\n");
			pthread_exit(&ret);
		}
	}

	while(1)
	{
		if(!(portopts & opt_udp))
		{
			ret = accept(fd, &addr, &addrsize);
			if(ret < 0)
			{
				fprintf(stderr, "Couldn't accept connection.\n");
				pthread_exit(&ret);
			}
		}
		else
		{
			ret = fd;
		}

		if(global_c.verbose) printf("--> Connection on port %i!\n", port);
		fflush(NULL);

#ifdef USE_TLS
		if(portopts & opt_tls)
		{
			ggz_tls_enable_fd(ret, GGZ_TLS_SERVER, GGZ_TLS_VERIFY_NONE);
		}
#endif

		if(!(portopts & opt_udp))
		{
			ret2 = getpeername(ret, &addr, &addrsize);
			if(ret2)
			{
				fprintf(stderr, "Peer name not found.\n");
				pthread_exit(&ret);
			}

			sa = *(struct sockaddr_in*)&addr;
			ipx[0] = ntohl(sa.sin_addr.s_addr);
			ipx[1] = (ipx[0] >> 24) & 0xFF;
			ipx[2] = (ipx[0] >> 16) & 0xFF;
			ipx[3] = (ipx[0] >> 8) & 0xFF;
			ipx[4] = (ipx[0]) & 0xFF;
			snprintf(ip, sizeof(ip), "%u.%u.%u.%u", ipx[1], ipx[2], ipx[3], ipx[4]);
			ret2 = ipban_check(ip);
			if(ret2 == 1)
			{
				fprintf(stderr, "%s is blocked!\n", ip);
				close(fd);
				pthread_exit(&ret);
			}
		}

		host = magic_node(&global_c, "", port, ipx[0]);
		destport = magic_port(&global_c, host, port);
		if(!destport) destsocket = magic_socket(&global_c, host, port);

		if(!host)
		{
			fprintf(stderr, "No more cluster nodes available for port %i.\n", port);
			close(ret);
			continue;
		}

		if(global_c.verbose)
		{
			if(!destport)
				printf("--> Connect to %s:%s\n", host, destsocket);
			else 
				printf("--> Connect to %s:%i\n", host, destport);
		}
		fflush(NULL);

		carg = (struct conn_t*)malloc(sizeof(struct conn_t));
		carg->fd = ret;
		carg->hostname = strdup(host);
		carg->port = port;
		carg->portopts = portopts;
		carg->socket = portsocket;
		carg->destportopts = destportopts;
		carg->destport = destport;
		if(destsocket)
			carg->destsocket = strdup(destsocket);
		else
			carg->destsocket = NULL;
		if(prelaunch)
			carg->prelaunch = strdup(prelaunch);
		else
			carg->prelaunch = NULL;
		if(pattern)
			carg->pattern = strdup(pattern);
		else
			carg->pattern = NULL;

		/*id = (pthread_t*)malloc(sizeof(pthread_t));*/
		pthread_attr_init(&attr);
		pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
		pthread_create(&id, NULL, &server_handle, carg);
		pthread_attr_destroy(&attr);
	}

	pthread_exit(&ret);
}

/* Start all subservers */
void server_start(conf c)
{
	int i, j;
	pthread_t id;
	pthread_attr_t attr;
	struct conn_t *carg;
	char service[10];

	if(!c.clusternum)
	{
		fprintf(stderr, "No clusters found.\n");
		exit(-1);
	}

	magic_initialize(&c);
	global_c = c;

	for(i = 0; i < c.clusternum; i++)
	{
		for(j = 0; j < c.cluster[i]->portsnum; j++)
		{
			carg = (struct conn_t*)malloc(sizeof(struct conn_t));
			carg->fd = -1;
			carg->hostname = NULL;
			carg->port = c.cluster[i]->ports[j];
			carg->portopts = c.cluster[i]->portopts[j];
			carg->socket = c.cluster[i]->sockets[j];
			carg->destport = -1;
			carg->destportopts = c.cluster[i]->destportopts;
			carg->destsocket = NULL;
			if(c.cluster[i]->prelaunch)
				carg->prelaunch = strdup(c.cluster[i]->prelaunch);
			else
				carg->prelaunch = NULL;
			if(c.cluster[i]->pattern)
				carg->pattern = strdup(c.cluster[i]->pattern);
			else
				carg->pattern = NULL;
			if(carg->pattern)
			{
				snprintf(service, sizeof(service), "%i", carg->port);
				filter_setup(carg->pattern, service);
			}

			/*id = (pthread_t*)malloc(sizeof(pthread_t));*/
			pthread_attr_init(&attr);
			pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
			pthread_create(&id, NULL, &server_sub, carg);
			pthread_attr_destroy(&attr);
		}
	}

	if(global_c.verbose) printf("Server started.\n");
	while(1)
		sleep(99999);
}

